package PureMessage::Sophos::SAVI::TFT;

use strict;
use warnings;
use Carp qw(croak);

use base qw(Exporter);
our @EXPORT_OK = qw(get_tft_mime_types 
                    get_tft_extensions
                    get_tft_filetypes
                    scan_part
                    scan_file
                    tft_part_finished
                    get_part_content_file
                    remove_part_content_file
                    get_all_tft_filetypes
                    get_tft_mime_types_mapping
                    get_mime_types_for_id
                    get_extensions_for_id
                    get_tft_savi
                    get_part_filenames
                    );

use PureMessage::Sophos::SAVI::TFT::Config qw(%tft_config);
use Sophos::SAVI qw(get_savi);
use Sophos::SAVI::Engine qw($NOTIFY_DEFAULT $NOTIFY_STOP);
use File::Basename;
use File::Path qw(mkpath);
use PureMessage::Version qw($PREFIX);

use constant NUM_OF_WARNING_LINES => 5;

my @warning_msgs;
my @tft_ids;
my @tft_info;
my $tft_filename;
my ($file_count, $class_count);
my $notify = $tft_config{inspect_archives}{enabled}
    ? $NOTIFY_DEFAULT : $NOTIFY_STOP;
my @filenames;

sub cb_onfilefound {
    $tft_filename = shift;
    if (defined $tft_filename) {
        $tft_filename = basename($tft_filename);
        push @filenames, $tft_filename if ($file_count);
    }
    $file_count++;
    return $NOTIFY_DEFAULT;
}

sub cb_onclassification {
    $class_count++;
    push @tft_ids, shift;
    return $notify;
}

sub cb_onclassificationex {
    my ($id, $group, $identity) = @_;

    # Arguments:
    # - $id       = the numeric id that was formerly returned by onclassification().
    #               For extended TFT types, this is always 0xffff.
    # - $identity = the unique identity name, eg: Microsoft Word-OPC
    # - $group    = the group, eg: Document

    # print "OnClassificationEx - START ($id / $group / $identity)\n";

    if (!$id or !$identity) {
        # print "OnClassificationEx - returning notify\n";
        return $notify;
    }
    # print "OnClassificationEx - NOT returning notify\n";

    my $extensions = [];
    my $mime_types = [];

    my $id_data = $tft_config{ids}{$identity};

    if (defined $id_data) {
	    $extensions = $id_data->{extensions} || [];
	    $mime_types = $id_data->{mime_types} || [];
    }
    # print "OnClassificationEx - adding (id='$id' / name='$name' / ext='$ext')\n";

    # Name exposed to the user is Group/Identity, eg: Document/Microsoft Word-OPC
    push @tft_info, {
            fileid      => $file_count,
            id          => $id,
            name        => "$group/$identity", 
            identity    => $identity,
            extensions  => $extensions,
            mime_types  => $mime_types,
            tft_filename=> $tft_filename,
    };

    return $notify;
}

our $tft_savi = undef;
sub get_tft_savi {
    return if defined $tft_savi;

    my $callbacks = {
        OnFileFound         => \&cb_onfilefound,
        OnClassification    => \&cb_onclassification,
        OnClassificationEx  => \&cb_onclassificationex,
    };

    $tft_config{backend}{Sophos}{MaxRecursionDepth} =
        $tft_config{inspect_archives}{max_recursion_depth}
            if defined $tft_config{inspect_archives}{max_recursion_depth};

    $tft_savi = get_savi($tft_config{backend}{Sophos},
                    notify       => $callbacks,
            );

    ($tft_savi = undef, croak 'cannot create Sophos::SAVI::Engine object!')
	unless defined $tft_savi;
    return $tft_savi;
}

sub scan_part {
    my $part = shift;
    @warning_msgs = ();
    local $SIG{__WARN__} = sub { 
        push @warning_msgs, $_[0]; 
        shift @warning_msgs if @warning_msgs > NUM_OF_WARNING_LINES 
    };
    $part->{tft_info} = scan_file($part);
    if (@warning_msgs) {
        $part->log('WARNING', $_) for @warning_msgs;
    }
}

sub _filter_by_type {
    my $tft_info = shift;

    # XXX If this list needs to be continually modified we probably want to
    # expose the values in the filetypes.conf.

    # If part contains 'PDF' remove all subsequent entries 'JavaScript'.
    my %exceptions = (
       'PDF' => ['JavaScript', 'XML', 'Unspecified Markup'],
    );

    # <Flatlist> -> <Split by Part>
    my @parts = ();
    my ($last_id, $idx) = (-1, -1);
    for (@{$tft_info}) {
       if ($last_id != $_->{fileid}) {
          $last_id = $_->{fileid};
          $idx++
       }
       push(@{$parts[$idx]}, $_);
    }

    # Filter -> <Flatlist>
    my @newparts = ();
    push(@newparts, _strip_subtypes($_, \%exceptions)) for (@parts);

    return @newparts;
}

sub _strip_subtypes {
    my ($part_info, $exceptions) = @_;

    return ()            unless $part_info;
    return @{$part_info} unless @{$part_info} > 1;

    my @file_types = @{$part_info};
    my $end = @file_types - 1;

    for (my $i = 0; $i < $end; $i++) {
        next unless $file_types[$i];
        (my $top_level_type = $file_types[$i]->{name}) =~ s#^.*/##;
        if (exists $exceptions->{$top_level_type}) {
           for (my $j = $i+1; $j < @file_types; $j++) {
               next unless $file_types[$j];
               (my $type = $file_types[$j]->{name}) =~ s#^.*/##;
               if (grep{$type eq $_} @{$exceptions->{$top_level_type}}) {
                  $file_types[$j] = undef;
               }
           }
        }
    }
    return grep { defined($_) } @file_types;
}

sub scan_file {
    my $part = shift;

    eval{$part->part_id();};
    if($@){
        $part = DummyObject->new($part);
    }

    get_tft_savi();

    # Initialize globals (populated by callbacks)
    @tft_ids = ();
    @tft_info = ();
    $file_count = $class_count = 0;
    @filenames = ();

   my $scan_context;
    my %property_hash;
    my $sav_ids;
    $scan_context = $tft_savi->set_scan_context('Genes/Extn/ProdVer','PMX:Email:3.9.1');
    $property_hash{'Genes/Extn/ProdVer'} = 1;

    if ( $part->part_id eq "1" ) {
        $scan_context = $tft_savi->set_scan_context('Email/MIME/Content-Type','message/rfc822');
        $property_hash{'Email/MIME/Content-Type'} = 1;
        $sav_ids = $tft_savi->scan_file($part->body_filename);
    }else{
        my $val= $part->header_decoded('Content-Disposition');
        if ( defined $val ) {
            $scan_context = $tft_savi->set_scan_context('Email/MIME/Content-Disposition',$val);
            $property_hash{'Email/MIME/Content-Disposition'} = 1;
        }
        $val= $part->header_decoded('Content-Type');
        if ( defined $val ) {
            $scan_context = $tft_savi->set_scan_context('Email/MIME/Content-Type',$val);
            $property_hash{'Email/MIME/Content-Type'} = 1;
        }
        $sav_ids = $tft_savi->scan_file($part->body_filename);
    }

    # delete properties set on scan context
    for my $property (keys %property_hash) {
        $tft_savi->delete_scan_property($property);
    }


    push @tft_ids, 0 unless $file_count == $class_count;

    my %extensions = ();
    my %mime_types = ();
    my %filetypes  = ();
    my %identities = ();
    my %filetype_mapping = ();

    # Old TFT functionality
    for (@tft_ids) {
        # Map ids to mime types
        my $mime_list = $tft_config{mapping}{sophos_type}->{"_$_"};
        if(defined $mime_list) {
            my @mime_list = (ref($mime_list) eq 'ARRAY') ? @$mime_list : ($mime_list);
            $mime_types{$_}++ for @mime_list;
        }
        # Map ids to extensions
        my $extension_list = $tft_config{mapping}{sophos_name}->{"_$_"};
        if(defined $extension_list) {
            my @extension_list = (ref($extension_list) eq 'ARRAY') ? @$extension_list : ($extension_list);
            $extensions{$_}++ for @extension_list;
        }
    }

    # New TFT functionality
    foreach my $info (_filter_by_type(\@tft_info)) {

        $filetypes{$info->{name}}++ if defined $info->{name};
        $identities{$info->{identity}}++ if defined $info->{identity};

        # Maps filename returned by TFT to the mime types of that file
        $filetype_mapping{$info->{tft_filename}}->{mime_types} ||= []; 
        $filetype_mapping{$info->{tft_filename}}->{extensions} ||= []; 
        $filetype_mapping{$info->{tft_filename}}->{filetypes} ||= []; 

        if (defined $info->{name}) {
            push(@{$filetype_mapping{$info->{tft_filename}}->{filetypes}}, $info->{name});
        }

        if ($info->{extensions} and @{$info->{extensions}}) {
            $extensions{$_}++ for @{$info->{extensions}};
            push(@{$filetype_mapping{$info->{tft_filename}}->{extensions}}, @{$info->{extensions}});
        }
        if ($info->{mime_types} and @{$info->{mime_types}}) {
            $mime_types{$_}++ for @{$info->{mime_types}};
            push(@{$filetype_mapping{$info->{tft_filename}}->{mime_types}}, @{$info->{mime_types}});
        }
    }
    
    my @extensions = sort keys %extensions;
    my @filetypes  = sort keys %filetypes;
    my @mime_types = sort keys %mime_types;
    my @identities = sort keys %identities;
    my @part_filenames = @filenames;

    return {
        extensions => \@extensions,
        filetypes  => \@filetypes,
        mimetypes  => \@mime_types,
        identities => \@identities,
        filetype_mapping => \%filetype_mapping,
        sav_ids => $sav_ids,
        filenames => \@part_filenames
    };
}

sub get_tft_extensions {
    my $part = shift;
    
    scan_part($part) if not tft_part_finished($part);

    return $part->{tft_info}->{extensions};
}

sub get_tft_filetypes {
    my $part = shift;

    scan_part($part) if not tft_part_finished($part);

    return $part->{tft_info}->{filetypes};
}

sub get_tft_mime_types {
    my $part = shift;

    scan_part($part) if not tft_part_finished($part);

    return $part->{tft_info}->{mimetypes};
}

sub get_tft_mime_types_mapping {
    my $part = shift;

    scan_part($part) if not tft_part_finished($part);

    return $part->{tft_info}->{filetype_mapping};
}

sub tft_part_finished { return exists shift->{tft_info} }

sub get_part_content_file {
    my $part = shift;
    if (not exists $part->{content_tmpfile}) {
        $part->{content_tmpfile} = $part->new_tempfile();
    }
    return $part->{content_tmpfile}->filename();
}

sub remove_part_content_file {
    my $part = shift;
    if (exists $part->{content_tmpfile}) {
        $part->{content_tmpfile}->release();
        delete $part->{content_tmpfile};
    }
}

sub get_all_tft_filetypes {
    get_tft_savi();
    return $tft_savi->get_tft_identities();
}

sub get_mime_types_for_id {
    my $id = shift;
    my $types = $tft_config{ids}{$id}{mime_types} || [];
    return $types;
}

sub get_extensions_for_id {
    my $id = shift;
    my $exts = $tft_config{ids}{$id}{extensions} || [];
    return $exts;
}

sub get_part_filenames {
    my $part = shift;

    scan_part($part) if not tft_part_finished($part);

    return $part->{tft_info}->{filenames};
}

1;

package DummyObject;

sub new{
        my $self = shift;
        my $file = shift;

        return bless{filename => $file}, $self;
}

sub part_id{
        return 1;
}

sub body_filename{
        return shift->{filename};
}
1;

